<'
# Copyright (c) 2011-2013 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Fri Mar  9 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]

lang c
'>
#include "autoconf/acheader.h"

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include "h2devLib.h"

#include "<"$comp">_control_task.h"
#include "<"$comp">_portlib.h"
#include "serialize.h"

/* --- local data ---------------------------------------------------------- */

enum {
  <"$comp">_PORTLIB_MIN_INDEX = 0,
<'foreach p [$component ports] {'>
  <"$comp">_<"[$p name]">_INDEX,
<'}'>
  <"$comp">_PORTLIB_MAX_INDEX
};

/* --- state port ---------------------------------------------------------- */

int
genom_state_<"$comp">_init(struct genom_<"$comp">_cntrlstr *cids)
{
  genom_state_component *data = &cids->ports.genom_state.h.buffer;
  genom_event s;

  /* task list */
<'
set max [[[dotgen types ::genom::state::component::task] type] length]
set tasks [lrange [$component tasks] 0 ${max}-1]
'>
  data->task._length = <"[llength $tasks]">;
<'foreach t $tasks {'>
  strncpy(data->task._buffer[<"[lsearch $tasks $t]">].name, "<"[$t name]">",
          sizeof(data->task._buffer[<"[lsearch $tasks $t]">].name));
  cids->tasks.<"[$t name]">.stateid = <"[lsearch $tasks $t]">;
<'}'>

  /* automatic versioning - nice idea borrowed from ROS */
<'
# one must filter out from the digest the 'metadata' port that is private
set digest [$component digest {{v} {
  if {[$v class] == "port" && [$v name] == "genom_metadata"} { return 0 }
  return 1
}}]
'>
  strncpy(data->digest, "<"$digest">", sizeof(data->digest));
  strncpy(data->date, "<"[clock format [clock sec]]">", sizeof(data->date));
  strncpy(data->version, "<"[$component version]">", sizeof(data->version));

  s = genom_<"$comp">_genom_state_write();
  if (s) {
    genom_log_warn(0, "cannot initialize state port: %s", s);
    return 1;
  }

  genom_log_debug("initialized outport genom_state");
  return 0;
}


/* --- metadata port ------------------------------------------------------- */

int
genom_metadata_<"$comp">_init(struct genom_<"$comp">_cntrlstr *cids)
{
  pocolibs_metadata_component *meta = &cids->ports.genom_metadata.h.buffer;
  size_t l, m;

  genom_sequence_reserve(&meta->services, <"[llength [$component services]]">);
  l = sizeof(meta->services._buffer->name);
  m = sizeof(meta->services._buffer->digest);
<'set i 0; foreach s [$component services] {'>
  meta->services._buffer[<"$i">].rqstid = <"$COMP">_<"[$s name]">_RQSTID;
  strncpy(meta->services._buffer[<"$i">].name, "<"[$s name]">", l);
  meta->services._buffer[<"$i">].name[l-1] = 0;
  strncpy(meta->services._buffer[<"$i">].digest, "<"[$s digest]">", m);
  meta->services._buffer[<"$i">].digest[m-1] = 0;
<'  incr i }'>
  meta->services._length = <"[llength [$component services]]">;

  genom_log_debug("initialized metadata port");
  return 0;
}

genom_event
genom_metadata_<"$comp">_fetch(const char *comp, pocolibs_metadata_component *meta)
{
  char name[H2_DEV_MAX_NAME];
  POSTER_ID id;
  char *data;
  STATUS s;
  ssize_t m;
  int n;

  n = snprintf(name, sizeof(name), "%s/genom_metadata", comp);
  if (n <= 0 || n >= sizeof(name)) {
    genom_log_warn(0, "port %s name too long", name);
    genom_log_warn(0, "port name length limited to %d characters",
                   H2_DEV_MAX_NAME);
    return genom_syserr(&(genom_syserr_detail){ .code = ENAMETOOLONG });
  }

  if (posterFind(name, &id) != OK) {
    genom_log_warn(1, "no such port %s", name);
    return genom_no_such_service();
  }

  s = posterTake(id, POSTER_READ);
  if (s != OK) {
    genom_log_warn(1, "cannot access metadata port %s", name);
    return genom_no_such_service();
  }

  data = posterAddr(id);
  m = -1;
  s = genom_deserialize_pocolibs_metadata_component(&data, &m, meta);
  if (s) {
    genom_log_warn(1, "cannot read metadata port %s contents", name);
    return genom_no_such_service();
  }
  posterGive(id);

  return 0;
}


<'foreach p [$component ports] {'>

/* <"[--- genom_${comp}_[$p name]_get_ph --------------------------------]"> */

static __inline__ struct genom_<"$comp">_<"[$p name]">_ph *
genom_<"$comp">_<"[$p name]">_get_ph(
<'  if {[$p kind] == "multiple"} {'>
  const char *id,
<'  }'>
  struct genom_<"$comp">_<"[$p name]">_port *p)
{
<'  if {[$p kind] == "multiple"} {'>
  int n;
  for(n = 0; n < p->nh; n++)
    if (p->h[n] && !strcmp(p->h[n]->name, id))
      return p->h[n];
  return NULL;
<'  } else {'>
  return &p->h;
<'  }'>
}


/* <"[--- genom_${comp}_[$p name]_data ----------------------------------]"> */

<"[[$p datatype] argument reference]">
genom_<"$comp">_<"[$p name]">_data(
<'  if {[$p kind] == "multiple"} {'>
  const char *id
<'  }'>
  )
{
  static struct genom_<"$comp">_<"[$p name]">_port *p;
  if (!p) p = &genom_<"$comp">_cids()->ports.<"[$p name]">;

  struct genom_<"$comp">_<"[$p name]">_ph *ph;
<'  if {[$p kind] == "multiple"} {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(id, p);
<'  } else {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(p);
<'  }'>
  if (!ph || !ph->id) return NULL;
  return <"[[$p datatype] pass reference {ph->buffer}]">;
}


/* <"[--- genom_${comp}_[$p name]_open ----------------------------------]"> */

genom_event
genom_<"$comp">_<"[$p name]">_open(
<'  if {"multiple" in [$p kind]} {'>
  const char *id
<'  }'>
  )
{
  static struct genom_<"$comp">_<"[$p name]">_port *p;
  if (!p) p = &genom_<"$comp">_cids()->ports.<"[$p name]">;

  struct genom_<"$comp">_<"[$p name]">_ph *ph;
  char name[H2_DEV_MAX_NAME];

<'  if {"out" in [$p dir]} {'>
  STATUS s;
  size_t l;
<'  }'>
  int n;

<'  if {"multiple" in [$p kind]} {'>
  n = snprintf(name, sizeof(name), "%s/<"[$p name]">/%s", genom_instance, id);
<'  } else {'>
  n = snprintf(name, sizeof(name), "%s/<"[$p name]">", genom_instance);
<'  }'>
  if (n <= 0 || n >= sizeof(name)) {
    genom_log_warn(0, "port <"[$p name]"> name too long");
    genom_log_warn(0, "port name length limited to %zu characters",
                   H2_DEV_MAX_NAME - strlen(genom_instance) - 2);
    return genom_syserr(&(genom_syserr_detail){ .code = ENAMETOOLONG });
  }

<'  if {"multiple" in [$p kind]} {'>
  for(n = 0; n < p->nh; n++)
    if (p->h[n] && !strcmp(p->h[n]->name, id))
      return genom_ok;
  for(n = 0; n < p->nh; n++)
    if (!p->h[n]) break;
  if (n >= p->nh) {
    struct genom_<"$comp">_<"[$p name]">_ph **pph;
    pph = realloc(p->h, (p->nh+1)*sizeof(*pph));
    if (!pph) return genom_syserr(&(genom_syserr_detail){ .code = ENOMEM });
    p->h = pph;
    n = p->nh++;
  }
  p->h[n] = malloc(sizeof(*ph));
  if (!p->h[n]) return genom_syserr(&(genom_syserr_detail){ .code = ENOMEM });

  ph = p->h[n];
  genom_tinit_<"[[$p datatype] mangle]">(
    <"[[$p datatype] pass reference ph->buffer]">);
  strncpy(ph->name, id, sizeof(ph->name));
<'  } else {'>
  n = 0;
  ph = &p->h;
<'  }'>
  ph->id = NULL;

<'  if {"out" in [$p dir]} {'>
  l = genom_maxserialen_<"[[$p datatype] mangle]">();

  s = posterCreate(name, (int)l, &ph->id);
  if (s == ERROR) {
    genom_log_warn(1, "cannot create outport %s", name);
<'    if {"multiple" in [$p kind]} {'>
    free(ph);
    p->h[n] = NULL;
<'    }'>
    return genom_syserr(&(genom_syserr_detail){ .code = ENOENT });
  }
  ph->size = l;
  genom_log_info("created outport %s", name);
<'  } else {'>
  genom_log_info("created inport %s", name);
<'  }'>
  return genom_ok;
}


/* <"[--- genom_${comp}_[$p name]_delete --------------------------------]"> */

genom_event
genom_<"$comp">_<"[$p name]">_close(
<'  if {"multiple" in [$p kind]} {'>
  const char *id
<'  }'>
  )
{
  static struct genom_<"$comp">_<"[$p name]">_port *p;
  if (!p) p = &genom_<"$comp">_cids()->ports.<"[$p name]">;

  struct genom_<"$comp">_<"[$p name]">_ph *ph;
<'  if {[$p kind] == "multiple"} {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(id, p);
<'  } else {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(p);
<'  }'>
  if (!ph || !ph->id)
    return genom_syserr(&(genom_syserr_detail){ .code = ENOENT });

<'  if {"out" in [$p dir]} {'>
  posterDelete(ph->id);
<'    if {"multiple" in [$p kind]} {'>
  genom_log_info("destroyed outport <"[$p name]">/%s", ph->name);
<'    } else {'>
  genom_log_info("destroyed outport <"[$p name]">");
<'    }'>
<'  } else {'>
<'    if {"multiple" in [$p kind]} {'>
  genom_log_info("destroyed inport <"[$p name]">/%s", ph->name);
<'    } else {'>
  genom_log_info("destroyed inport <"[$p name]">");
<'    }'>
<'  }'>
  ph->id = NULL;
  return genom_ok;
}

void
genom_<"$comp">_<"[$p name]">_delete(void)
{
<'  if {"multiple" in [$p kind]} {'>
  static struct genom_<"$comp">_<"[$p name]">_port *p;
  if (!p) p = &genom_<"$comp">_cids()->ports.<"[$p name]">;

  int n;
  for(n = 0; n < p->nh; n++)
    if (p->h[n]) {
      genom_<"$comp">_<"[$p name]">_close(p->h[n]->name);
      p->h[n] = NULL;
    }
<'  } else {'>
  genom_<"$comp">_<"[$p name]">_close();
<'  }'>
}

<'}'>

<'foreach p [$component ports in] {'>

/* <"[--- genom_${comp}_[$p name]_connect -------------------------------]"> */

genom_event
genom_<"$comp">_<"[$p name]">_connect(
<'  if {"multiple" in [$p kind]} {'>
  const char *id,
<'  }'>
  const char *name, struct genom_<"$comp">_cntrlstr *cids)
{
  struct genom_<"$comp">_<"[$p name]">_port *p = &cids->ports.<"[$p name]">;
  struct genom_<"$comp">_<"[$p name]">_ph *ph;
  POSTER_ID pid;

<'  if {[$p kind] == "multiple"} {'>
  genom_event s = genom_<"$comp">_<"[$p name]">_open(id);
  if (s) return s;
  ph = genom_<"$comp">_<"[$p name]">_get_ph(id, p);
<'  } else {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(p);
<'  }'>

  if (!ph) return genom_no_such_inport();

  if (posterFind(name, &pid) != OK) {
    genom_log_warn(1, "no such port %s", name);
    return genom_no_such_outport();
  }
  if (posterTake(pid, POSTER_READ) != OK) {
    genom_log_warn(1, "cannot connect port %s", name);
    return genom_port_io();
  }
  posterGive(pid);

  ph->id = pid;
  return genom_ok;
}


/* <"[--- genom_${comp}_[$p name]_read ----------------------------------]"> */

genom_event
genom_<"$comp">_<"[$p name]">_read(
<'  if {"multiple" in [$p kind]} {'>
  const char *id
<'  }'>
  )
{
  static struct genom_<"$comp">_<"[$p name]">_port *p;
  if (!p) p = &genom_<"$comp">_cids()->ports.<"[$p name]">;
  STATUS s;
  char *b;
  ssize_t max;

  struct genom_<"$comp">_<"[$p name]">_ph *ph;
<'  if {[$p kind] == "multiple"} {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(id, p);
<'  } else {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(p);
<'  }'>
  if (!ph || !ph->id) return genom_no_such_inport();

  s = posterTake(ph->id, POSTER_READ);
  if (s != OK) {
<'  if {"multiple" in [$p kind]} {'>
    genom_log_warn(1, "cannot access inport <"[$p name]">/%s", ph->name);
<'  } else {'>
    genom_log_warn(1, "cannot access inport <"[$p name]">");
<'  }'>
    return genom_port_io();
  }

  b = posterAddr(ph->id);
  max = -1;
  s = genom_deserialize_<"[[$p datatype] mangle]">(
    &b, &max, <"[[$p datatype] pass value ph->buffer]">);
  posterGive(ph->id);
  if (s) {
<'  if {"multiple" in [$p kind]} {'>
    genom_log_warn(0, "cannot read inport <"[$p name]">/%s contents", ph->name);
<'  } else {'>
    genom_log_warn(0, "cannot read inport <"[$p name]"> contents");
<'  }'>
    return genom_serialization();
  }

  return genom_ok;
}

<'}'>

<'foreach p [$component ports out] {'>

/* <"[--- genom_${comp}_[$p name]_write ---------------------------------]"> */

genom_event
genom_<"$comp">_<"[$p name]">_write(
<'  if {"multiple" in [$p kind]} {'>
  const char *id
<'  }'>
  )
{
  static struct genom_<"$comp">_<"[$p name]">_port *p;
  if (!p) p = &genom_<"$comp">_cids()->ports.<"[$p name]">;
  STATUS s;
  char *b;
  size_t l;

  struct genom_<"$comp">_<"[$p name]">_ph *ph;
<'  if {[$p kind] == "multiple"} {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(id, p);
<'  } else {'>
  ph = genom_<"$comp">_<"[$p name]">_get_ph(p);
<'  }'>
  if (!ph || !ph->id) return genom_no_such_outport();

  l = genom_serialen_<"[[$p datatype] mangle]">(
    <"[[$p datatype] pass value ph->buffer]">);
  if (l > ph->size) {
    s = posterIoctl(ph->id, FIO_RESIZE, &l);
    if (s != OK) {
<'  if {"multiple" in [$p kind]} {'>
      genom_log_warn(1, "cannot resize outport <"[$p name]">/%s to %zu bytes",
                     ph->name, l);
<'  } else {'>
      genom_log_warn(1, "cannot resize outport <"[$p name]"> to %zu bytes", l);
<'  }'>
      return genom_serialization();
    }
    ph->size = l;
  }

  s = posterTake(ph->id, POSTER_WRITE);
  if (s != OK) {
<'  if {"multiple" in [$p kind]} {'>
    genom_log_warn(1, "cannot access outport <"[$p name]">/%s", ph->name);
<'  } else {'>
    genom_log_warn(1, "cannot access outport <"[$p name]">");
<'  }'>
    return genom_port_io();
  }

  b = posterAddr(ph->id);
  genom_serialize_<"[[$p datatype] mangle]">(
    &b, <"[[$p datatype] pass value ph->buffer]">);
  s = posterGive(ph->id);
  if (s != OK) {
<'  if {"multiple" in [$p kind]} {'>
    genom_log_warn(1, "cannot flush outport <"[$p name]">/%s", ph->name);
<'  } else {'>
    genom_log_warn(1, "cannot flush outport <"[$p name]">");
<'  }'>
    return genom_port_io();
  }

  if (l < ph->size) {
    s = posterIoctl(ph->id, FIO_RESIZE, &l);
    if (s != OK)
<'  if {"multiple" in [$p kind]} {'>
      genom_log_warn(1, "cannot resize outport <"[$p name]">/%s to %zu bytes",
                     ph->name, l);
<'  } else {'>
      genom_log_warn(1, "cannot resize outport <"[$p name]"> to %zu bytes", l);
<'  }'>
    else
      ph->size = l;
  }

  return genom_ok;
}

<'}'>

/* --- end of file --------------------------------------------------------- */
